"""File carving test data generator."""

import woodblock.datagen
import woodblock.errors
import woodblock.file
import woodblock.fragments
import woodblock.image
import woodblock.random
import woodblock.scenario
import woodblock.utils



"""This module contains the Woodblock errors/exceptions."""


class WoodblockError(Exception):
    """Base class for all Woodblock errors."""


class InvalidFragmentationPointError(WoodblockError):
    """This error indicates an invalid fragmentation point."""


class ImageConfigError(WoodblockError):
    """This error is raised when an invalid image configuration file is used."""



"""Utility functions."""

import hashlib
import pathlib
from functools import lru_cache


@lru_cache(maxsize=1024)
def hash_file(path: pathlib.Path) -> str:
    """Return the SHA-256 hash of a file as hexadecimal string.

    The results of this functions are cached in order to avoid multiple computations for the same file.

    Args:
        path: The path of the file to hash
    """
    sha256 = hashlib.sha256()
    with path.open('rb') as file_handle:
        for chunk in iter(lambda: file_handle.read(1048576), b''):
            sha256.update(chunk)
    return sha256.hexdigest()


def get_file_list(path: pathlib.Path, min_size: int = 0) -> tuple:
    """Return all files from the given `path`.

    This function returns a list of all files paths relative to `path`. The resulting file list will be sorted so that
    the order of the files will always be the same for identical directory/file hierarchies.

    Note that README files (and variations such as README.md, ReadMe etc.) are NOT included in the result list.

    Args:
        path: The directory relative to the file corpus from which the file should be chosen.
        min_size: Minimal file size of the selected files.
    """
    basedir = pathlib.Path(path)
    return tuple(
        sorted(f for f in basedir.glob('**/*') if f.is_file() and not _is_readme(f) and f.stat().st_size >= min_size))


def _is_readme(path) -> bool:
    name = path.name.lower()
    return name.startswith('readme')



"""This module takes care about the randomness and reproducibility of the generated data sets."""

import random

import numpy as np


class Randomness:
    """This class acts as the main random number generator.

    ``Randomness`` is basically just a wrapper for the random modules of the Python core and NumPy. All it does is
    initializing the seed of both modules so that the user doesn't have to do this.

    If no seed is set by the user, the seed will be randomly generated.
    """

    def __init__(self):
        self._seed = random.randint(0, 2 ** 32 - 1)  # nosec
        self.seed(self._seed)

    def seed(self, random_seed: int):  # pylint: disable=redefined-outer-name
        """Set the seed.

        Args:
            random_seed: The seed to use.
        """
        self._seed = random_seed
        random.seed(self._seed)
        np.random.seed(self._seed)

    def get_seed(self) -> int:
        """Return the random seed."""
        return self._seed


_RANDOM = Randomness()

seed = _RANDOM.seed  # pylint: disable=invalid-name

get_seed = _RANDOM.get_seed  # pylint: disable=invalid-name


class RandomBytes:  # pylint: disable=too-few-public-methods
    """This class can be used to generate random bytes."""

    def __init__(self):
        self._rng = np.random.RandomState()  # pylint: disable=no-member

    def bytes(self, size: int):
        """Return ``size`` random bytes.

        Args:
            size: The number of bytes to return.
        """
        return self._rng.bytes(size)

    def seed(self, random_seed: int):
        """Set the seed for the RNG.

        Args:
            seed: The seed to use.
        """
        self._rng.seed(random_seed)



import pathlib
import sys

import click

import woodblock

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
def main():
    pass


@main.command(name='generate')
@click.argument('config', type=click.Path(exists=True))
@click.argument('image', type=click.Path())
def generate_image(config, image):
    """Generate an image based on the given configuration file.

    \b
    CONFIG is the path to the configuration file to use.
    IMAGE  is the output path of the generated image."""
    img = woodblock.image.Image.from_config(pathlib.Path(config))
    img.write(pathlib.Path(image))


if __name__ == '__main__':
    sys.exit(main())



"""This module contains the Image class."""
import configparser
import json
import pathlib
import random
from collections import defaultdict
from operator import itemgetter

from multimethod import multimethod

import woodblock.datagen
import woodblock.file
from woodblock.errors import ImageConfigError
from woodblock.scenario import Scenario


class Image:
    """The Image class represents a carving test image.

    An image contains a sequence of `Scenario` instances. An image has a fixed block size. All blocks smaller than the
    block size will be padded with data generated by a configurable data generator.

    Args:
        block_size: Block size of the image.
        padding_generator: A data generator used to generate padding.
    """

    def __init__(self, block_size: int = 512, padding_generator=woodblock.datagen.Random()):
        self._block_size = block_size
        self._scenarios = list()
        self._generate_padding = padding_generator

    def add(self, scenario):
        """Add a ``Scenario`` to the image."""
        self._scenarios.append(scenario)

    @staticmethod
    def from_config(path: pathlib.Path):
        """Create an Image instance based on a configuration file.

        Args:
            path: The path to the configuration file.
        """
        config = configparser.RawConfigParser()
        with path.open('r') as config_handle:
            try:
                config.read_file(config_handle)
            except configparser.Error as err:
                raise ImageConfigError(err.message)
        general = _parse_general_section(config)
        woodblock.file.corpus(path.absolute().parent / general['corpus'])
        if 'seed' in general:
            woodblock.random.seed(general['seed'])
        num_filler_blocks = (general['min filler blocks'], general['max filler blocks'])
        image = Image()
        for section in config.sections():
            if section != 'general':
                scenario = _parse_scenario_section(section_name=section, section=config[section],
                                                   num_filler_blocks=num_filler_blocks,
                                                   block_size=general['block size'])
                image.add(scenario)
        return image

    @multimethod
    def write(self, path: pathlib.Path):
        """Write the image to disk.

        This method write the image to the specified ``path``. Moreover, it also writes the image metadata to disk. The
        metadata file will be ``path`` with the ".json" extension. E.g. if ``path`` is "test-image.dd" then the actual
        image will be in "test-image.dd" and the metadata will be in "test-image.dd.json".

        Args:
            path: The output path.
        """
        with path.open('wb') as file_handle:
            self.write(file_handle)
        self._write_metadata(path)

    @multimethod
    def write(self, image_handle):  # pylint: disable=function-redefined
        """Write the image to disk.

        This method write the image to the specified ``path``. Note that no metadata is written. You have to call
        metadata yourself if you need it.

        Args:
            image_handle: A ``.write()``-supporting file-like object.
        """
        for scenario in self._scenarios:
            self._write_scenario(image_handle, scenario)

    @property
    def metadata(self):
        """Return the image metadata."""
        meta = {'block_size': self._block_size, 'seed': woodblock.random.get_seed(),
                'corpus': str(woodblock.file.get_corpus()),
                'scenarios': [s.metadata for s in self._scenarios]}
        self._update_metadata_with_image_offsets(meta, self._compute_image_offsets())
        return meta

    def _write_metadata(self, image_path: pathlib.Path):
        metadata_path = pathlib.Path('.'.join((str(image_path.absolute()), 'json')))
        with metadata_path.open('w') as file_handle:
            json.dump(self.metadata, file_handle)

    def _write_scenario(self, image, scenario):
        for fragment in scenario:
            for chunk in fragment:
                image.write(chunk)
            self._pad_fragment(image, fragment.size)

    def _pad_fragment(self, image, fragment_size):
        remainder = fragment_size % self._block_size
        if remainder != 0:
            padding_size = self._block_size - remainder
            image.write(self._generate_padding(padding_size))

    def _compute_image_offsets(self):
        image_offsets = defaultdict(dict)
        current_offset = 0
        for scenario in self._scenarios:
            for frag in scenario:
                frag_meta = frag.metadata
                frag_size = frag_meta['fragment']['size']
                end_offset = current_offset + frag_size
                image_offsets[frag_meta['file']['id']][frag_meta['fragment']['number']] = {'start': current_offset,
                                                                                           'end': end_offset}
                current_offset += frag_size
                if current_offset % self._block_size != 0:
                    current_offset += self._block_size - (current_offset % self._block_size)
        return image_offsets

    @staticmethod
    def _update_metadata_with_image_offsets(meta, image_offsets):  # pylint: disable=invalid-name
        for scenario_meta in meta['scenarios']:
            for file_meta in scenario_meta['files']:
                file_id = file_meta['original']['id']
                for frag_meta in file_meta['fragments']:
                    frag_meta['image_offsets'] = image_offsets[file_id][frag_meta['number']]


def _parse_general_section(config: dict) -> dict:
    if 'general' not in config:
        raise ImageConfigError('Mandatory "general" section is not present.')
    section = config['general']
    if 'corpus' not in section:
        raise ImageConfigError('Mandatory "corpus" key in "general" section is not present.')
    general = dict()
    general['corpus'] = section['corpus']
    if 'block size' in section:
        general['block size'] = int(section['block size'])
    else:
        general['block size'] = 512
    if 'seed' in section:
        general['seed'] = int(section['seed'])
    general['min filler blocks'] = _get_min_filler_fragment_blocks(section) or 1
    general['max filler blocks'] = _get_max_filler_fragment_blocks(section) or 10
    return general


def _get_min_filler_fragment_blocks(section):
    return _get_number_of_blocks(section, 'min')


def _get_max_filler_fragment_blocks(section):
    return _get_number_of_blocks(section, 'max')


def _get_number_of_blocks(section: dict, min_or_max: str):
    if '{} filler blocks'.format(min_or_max) in section:
        blocks = int(section['{} filler blocks'.format(min_or_max)])
        if blocks < 1:
            raise ImageConfigError('Invalid value for "{} filler blocks": {}.'.format(min_or_max, blocks))
        return blocks
    return None


def _parse_scenario_section(section_name: str, section: dict, num_filler_blocks: tuple, block_size: int):
    try:
        layout = _get_layout_type(section)
    except KeyError:
        raise ImageConfigError('Section [{}] contains no "layout" key.'.format(section_name))

    if layout == 'fragment-sequence':
        return _parse_fragment_sequence_layout(section_name, section, num_filler_blocks, block_size)
    if layout == 'intertwine':
        return _parse_intertwine_layout(section_name, section, block_size)
    raise ImageConfigError('Unsupported layout type: ')


def _get_layout_type(section: dict) -> str:
    layout = section['layout'].lower()
    if layout == 'intertwine':
        return 'intertwine'
    return 'fragment-sequence'


def _parse_intertwine_layout(section_name: str, section: dict, block_size: int):
    scenario = Scenario(section_name)
    try:
        num_files = section['num files']
    except KeyError:
        raise ImageConfigError(
            'Section [{}] has an intertwined layout but does not specify "num files".'.format(section_name))
    num_files = int(num_files)
    min_frags, max_frags = _parse_frags_nums(section_name, section)
    scenario.add(woodblock.file.intertwine_randomly(number_of_files=num_files, block_size=block_size,
                                                    min_fragments=min_frags, max_fragments=max_frags))
    return scenario


def _parse_fragment_sequence_layout(section_name: str, section: dict, num_filler_blocks: tuple, block_size: int):
    scenario = Scenario(section_name)
    files = dict()
    for key, value in section.items():
        if key.startswith('file'):
            file_number = int(key[4:])
            if file_number in files:
                files[file_number]['path'] = value
            else:
                files[file_number] = {'frags': None, 'path': value}
        elif key.startswith('frags_file') or key.startswith('frags file'):
            file_number = int(key[10:])
            if file_number in files:
                files[file_number]['frags'] = int(value)
            else:
                files[file_number] = {'frags': int(value), 'path': None}
    _assert_each_file_has_a_defined_num_of_frags(files, section_name)
    file_fragments = _create_file_fragments(files)
    layout = _parse_layout_line(section['layout'])

    min_filler_blocks = _get_min_filler_fragment_blocks(section) or num_filler_blocks[0]
    max_filler_blocks = _get_max_filler_fragment_blocks(section) or num_filler_blocks[1]
    if min_filler_blocks < 1 or max_filler_blocks < 1 or min_filler_blocks > max_filler_blocks:
        raise ImageConfigError(
            'Invalid min/max number of fillers blocks: min={}, max={}'.format(min_filler_blocks, max_filler_blocks))
    for fragment in layout:
        if fragment['type'] == 'file':
            scenario.add(file_fragments[fragment['file_num']][fragment['frag_num']])
        elif fragment['type'] == 'random':
            size = _get_filler_fragment_size(min_filler_blocks, max_filler_blocks, block_size)
            scenario.add(woodblock.fragments.RandomDataFragment(size))
        elif fragment['type'] == 'zeroes':
            size = _get_filler_fragment_size(min_filler_blocks, max_filler_blocks, block_size)
            scenario.add(woodblock.fragments.ZeroesFragment(size))
    return scenario


def _parse_frags_nums(section_name: str, section: dict) -> tuple:
    min_frags = 1
    max_frags = 4
    if 'min frags' in section:
        try:
            min_frags = int(section['min frags'])
        except ValueError:
            raise ImageConfigError('"min frags" has to be an integer > 0 in section [{}]'.format(section_name))
        else:
            if min_frags < 1:
                raise ImageConfigError('"min frags" has to be an integer > 0 in section [{}]'.format(section_name))
    if 'max frags' in section:
        try:
            max_frags = int(section['max frags'])
        except ValueError:
            raise ImageConfigError('"max frags" has to be an integer > 0 in section [{}]'.format(section_name))
        else:
            if max_frags < 1:
                raise ImageConfigError('"max frags" has to be an integer > 0 in section [{}]'.format(section_name))
    if min_frags > max_frags:
        raise ImageConfigError('"min frags" has to be <= "max frags" in section [{}]'.format(section_name))
    return min_frags, max_frags


def _get_filler_fragment_size(min_blocks: int, max_blocks: int, block_size: int) -> int:
    return random.randint(min_blocks, max_blocks) * block_size  # nosec


def _create_file_fragments(files):
    fragments = defaultdict(dict)
    for file_num in files:
        num_frags = files[file_num]['frags']
        file_path = files[file_num]['path']
        if file_path is not None:
            if (woodblock.file.get_corpus() / file_path).is_dir():
                frags = woodblock.file.draw_fragmented_files(file_path, number_of_files=1, min_fragments=num_frags,
                                                             max_fragments=num_frags)[0]
            else:
                frags = woodblock.file.File(file_path).fragment_randomly(num_fragments=num_frags)
        else:
            frags = woodblock.file.draw_fragmented_files(number_of_files=1, min_fragments=num_frags,
                                                         max_fragments=num_frags)[0]
        fragments[file_num] = {i + 1: f for i, f in enumerate(frags)}
    return fragments


def _assert_each_file_has_a_defined_num_of_frags(files, section):  # pylint: disable=invalid-name
    for num, file in files.items():
        if file['frags'] is None:
            raise ImageConfigError('No "frags file{} definition is missing in section "[{}]".'.format(num, section))


def _parse_layout_line(line: str) -> list:
    layout = list()
    for token in (x.strip() for x in line.split(',')):
        token = token.lower()
        if token == 'r':  # nosec
            layout.append({'type': 'random'})
        elif token == 'z':  # nosec
            layout.append({'type': 'zeroes'})
        else:
            file_num, frag_num = _parse_file_definition(token)
            layout.append({'type': 'file', 'file_num': int(file_num), 'frag_num': int(frag_num)})
    return layout


def _parse_file_definition(token: str):
    if '.' in token:
        sep = '.'
    elif '-' in token:
        sep = '-'
    else:
        raise ImageConfigError('No valid file/fragment separator found. Use "-" or ".".')
    return token.split(sep)


class ImageLogParser:
    def __init__(self, fp):
        self._log = json.load(fp)

    def get_fragment_order(self):
        fragments = list()
        for scenario in self._log['scenarios']:
            for file in scenario['files']:
                fragments.extend((scenario['name'], file['original']['id'], file['original']['path'],
                                  frag['number'], frag['file_offsets']['start'], frag['file_offsets']['end'],
                                  frag['image_offsets']['start'], frag['image_offsets']['end'])
                                 for frag in file['fragments'])
        return tuple(sorted(fragments, key=itemgetter(6)))



"""This module contains file carving scenario related classes and functions."""

from multimethod import multimethod

import woodblock.fragments


class Scenario(list):
    """This class represents a file carving scenario.

    A scenario contains fragments in a certain order.

    Args:
        name: The name of the scenario.
    """

    def __init__(self, name: str):
        list.__init__([])
        self.name = name

    @multimethod
    def add(self, fragment: woodblock.fragments.FillerFragment):
        """Add a filler fragment to the scenario.

        Args:
            fragment: The fragment to be added.
        """
        self.append(fragment)

    @multimethod
    def add(self, fragment: woodblock.fragments.FileFragment):  # pylint: disable=function-redefined
        """Add a file fragment to the scenario.

        Args:
            fragment: The fragment to be added.
        """
        self.append(fragment)

    @multimethod
    def add(self, fragments: list):  # pylint: disable=function-redefined
        """Add a list of fragments to the scenario.

        Args:
            fragments: The list of fragments to be added.
        """
        self._add_from_iterable(fragments)

    @multimethod
    def add(self, fragments: tuple):  # pylint: disable=function-redefined
        """Add a tuple of fragments to the scenario.

        Args:
            fragments: The tuple of fragments to be added.
        """
        self._add_from_iterable(fragments)

    def _add_from_iterable(self, iterable):
        self.extend(iterable)

    @property
    def metadata(self) -> dict:
        """Return the scenario metadata."""
        meta = {'name': self.name, 'files': list()}
        files = dict()
        for frag in self:
            frag_meta = frag.metadata
            file_id = frag_meta['file']['id']
            if file_id not in files:
                files[file_id] = {'original': frag_meta['file'], 'fragments': list()}
            files[file_id]['fragments'].append(frag_meta['fragment'])
        meta['files'] = list(files.values())
        self._sort_fragments_by_number(meta)
        return meta

    @staticmethod
    def _sort_fragments_by_number(meta):
        for file in meta['files']:
            file['fragments'] = list(sorted(file['fragments'], key=lambda x: x['number']))



"""File related classes and functions."""
import itertools
import math
import pathlib
import random
import typing
from operator import attrgetter
from uuid import uuid4

import numpy as np

import woodblock.utils
from woodblock.errors import WoodblockError, InvalidFragmentationPointError
from woodblock.fragments import FileFragment

_CORPUS = None


def corpus(path):
    """Set the file corpus.

    Args:
        path: Path to the corpus directory.
    """
    global _CORPUS  # pylint: disable=global-statement
    if isinstance(path, str):
        _CORPUS = pathlib.Path(path)
    elif isinstance(path, pathlib.Path):
        _CORPUS = path
    elif path is None:
        _CORPUS = None
    else:
        raise WoodblockError('Unsupported object type for corpus.')


def get_corpus() -> pathlib.Path:
    """Return the path to the test file corpus."""
    if _CORPUS is None:
        raise WoodblockError('No corpus specified.')
    return _CORPUS


class File:
    """This class represents an actual file of the test file corpus.

    Args:
        path: Path to the file on disk.
    """

    def __init__(self, path: pathlib.Path):
        self._path = get_corpus() / path
        if not self.path.exists():
            raise FileNotFoundError(path.name)
        self._size = self._path.stat().st_size
        self._id = uuid4().hex
        self._hash = None
        if self._size < 1:
            raise WoodblockError('File contains no data!')

    def __str__(self):
        return ': '.join((self.__repr__(), self._path))  # pragma: no cover

    @property
    def hash(self) -> str:
        """Return the SHA-256 hash of the file as hexadecimal string."""
        if self._hash is None:
            self._hash = woodblock.utils.hash_file(self._path)
        return self._hash

    @property
    def id(self) -> str:  # pylint: disable=invalid-name
        """Return the ID of the file."""
        return self._id

    @property
    def path(self) -> pathlib.Path:
        """Return the path of the file relative to the corpus path."""
        return self._path

    @property
    def size(self) -> int:
        """Return the file size."""
        return self._size

    def as_fragment(self):
        """Return the file as a single fragment."""
        return FileFragment(file=self, fragment_number=1, start_offset=0, end_offset=self._size)

    def max_fragments(self, block_size):
        """Return the maximal number of fragments which can be created for a given block size."""
        if block_size < 1:
            raise WoodblockError('Block size has to be at least 1.')
        return math.ceil(self.size / block_size)

    def fragment(self, fragmentation_points: typing.Sequence, block_size: int = 512) -> list:
        """Fragment the file at the given ``fragmentation_points`` using the given ``block_size``.

        This method fragments the current file at the specified ``fragmentation_points``. The fragmentation points are
        multiplied with the given ``block_size`` in order to compute the actual fragmentation offsets.

        Args:
            fragmentation_points: A sequence of fragmentation points.
            block_size: The block size to use.
        """
        if not fragmentation_points:
            return [self.as_fragment(), ]
        self._validate_fragmentation_points(fragmentation_points, block_size)
        fragments = list()
        frag_points = sorted(fragmentation_points)
        fragment_number = 2
        len_frag_points = len(frag_points)
        for i in range(len_frag_points):
            if i == 0:
                fragments.append(FileFragment(self, 1, 0, frag_points[0] * block_size))
            if i < len_frag_points - 1:
                fragments.append(
                    FileFragment(self, fragment_number, frag_points[i] * block_size, frag_points[i + 1] * block_size))
            if i == len_frag_points - 1:
                fragments.append(
                    FileFragment(self, len_frag_points + 1, frag_points[i] * block_size, self._size))
            fragment_number += 1
        return fragments

    def fragment_randomly(self, num_fragments: int = None, block_size: int = 512) -> list:
        """Fragment the file at random fragmentation points.

        This method fragments the current file into ``num_fragments`` fragments. The fragmentation points are chosen
        randomly. If `num_fragments` is None, then the number of fragments is chosen randomly between 1 and the maximum
        number of fragments for the given `block_size`.

        Args:
            num_fragments: Number of fragments to create.
            block_size: Block size to use.
        """
        blocks_in_file = math.floor(self._size / block_size)
        if num_fragments is None:
            num_fragments = random.randint(1, blocks_in_file)  # nosec
        if self._size < block_size:
            raise InvalidFragmentationPointError(
                'File "{}" is too small to be fragmented with a block size of {}!'.format(self.path, block_size))
        if num_fragments < 1:
            raise InvalidFragmentationPointError('Number of fragments has to be at least 1.')
        if num_fragments > blocks_in_file and not (
                num_fragments == blocks_in_file + 1 and self._size % block_size != 0):
            raise InvalidFragmentationPointError('Number of fragments is too large.')
        if num_fragments == 1:
            return [self.as_fragment(), ]
        max_frag_point = blocks_in_file
        if self._size % block_size != 0:
            max_frag_point += 1
        return self.fragment(random.sample(range(1, max_frag_point), num_fragments - 1), block_size)

    def fragment_evenly(self, num_fragments: int, block_size: int = 512) -> list:
        """Fragment the file evenly into ``num_fragment`` fragments.

        This method fragments the current file into ``num_fragments`` fragments. The fragmentation points are chosen
        so that each fragment will be of the same size (if possible).

        Args:
            num_fragments: Number of fragments to create.
            block_size: Block size to use.
        """
        blocks_in_file = math.floor(self._size / block_size)
        if self._size < block_size:
            raise InvalidFragmentationPointError(
                'File is too small to be fragmented with a block size of {}!'.format(block_size))
        if num_fragments < 1:
            raise InvalidFragmentationPointError('Number of fragments has to be at least 1.')
        if num_fragments > blocks_in_file and not (
                num_fragments == blocks_in_file + 1 and self._size % block_size != 0):
            raise InvalidFragmentationPointError('Number of fragments is too large.')
        if num_fragments == 1:
            return [self.as_fragment(), ]

        if num_fragments == blocks_in_file + 1 and self._size % block_size != 0:
            frag_points = range(1, blocks_in_file + 1)
        elif num_fragments == blocks_in_file:
            frag_points = range(1, blocks_in_file)
        else:
            frag_points = tuple(
                int(x[-1]) for x in np.array_split(np.array(range(1, blocks_in_file + 1)), num_fragments))[:-1]
        return self.fragment(frag_points, block_size)

    def _validate_fragmentation_points(self, fragmentation_points, block_size):
        if self._size < block_size:
            raise InvalidFragmentationPointError(
                'File is too small to be fragmented with a block size of {}!'.format(block_size))
        if not all(isinstance(x, int) for x in fragmentation_points):
            raise InvalidFragmentationPointError('Fragmentation points have to be integers.')
        if 0 in fragmentation_points:
            raise InvalidFragmentationPointError('0 is not a valid fragmentation point.')
        if not all(f < math.ceil(self._size / block_size) for f in fragmentation_points):
            raise InvalidFragmentationPointError('Fragmentation point is too large.')
        if not all(f > 0 for f in fragmentation_points):
            raise InvalidFragmentationPointError('Fragmentation points have to be positive integers.')
        if len(set(fragmentation_points)) < len(fragmentation_points):
            raise InvalidFragmentationPointError('Duplicate fragmentation points are not allowed.')


def draw_files(path: pathlib.Path = None, number_of_files: int = 1, unique: bool = False, min_size: int = 0) -> list:
    """Choose random files from the file corpus.

    If `path` is None, the complete corpus will be considered. If it set to a path relative to the corpus, then only
    files in this directory (and its subdirectories) are considered.

    Args:
        path: The directory relative to the file corpus from which the file should be chosen.
        number_of_files: The number of files to draw.
        unique: If set to True, the resulting list will contain no duplicates.
        min_size: Minimal file size of the selected files.
    """
    if number_of_files < 1:
        raise WoodblockError('Number of files has to be at least 1.')
    start_path = get_corpus()
    if path is not None:
        start_path = start_path / path
    candidates = woodblock.utils.get_file_list(start_path, min_size)
    if not candidates:
        raise WoodblockError('Given path does not contain enough files with a minimal size of {}.'.format(min_size))
    if unique:
        try:
            return list(File(f) for f in random.sample(candidates, k=number_of_files))
        except ValueError:
            raise WoodblockError('Not enough unique files to choose from.')
    else:
        return list(File(f) for f in random.choices(candidates, k=number_of_files))


def draw_fragmented_files(path: pathlib.Path = None, number_of_files: int = 1, block_size: int = 512,
                          min_fragments: int = 1, max_fragments: int = 4) -> list:
    """Choose random files from ``path`` and fragment them randomly.

    ``draw_fragmented_files`` chooses ``number_of_files`` files from the given ``path`` and fragments them at random
    fragmentation points. The number of fragments per file will be between ``min_fragments`` and ``max_fragments``
    (both numbers included, i.e. min_fragments <= number of fragments <= max_fragments).

    The result is a list of fragment lists, e.g. [ [f1.1, f1.2, f1.3], [f2.1, ], [f3.1, f3.2] ].

    If `path` is None, the complete corpus will be considered. If it set to a path relative to the corpus, then only
    files in this directory (it its subdirectories) are considered.

    Note that there is no guarantee that a file is not chosen more than once.

    Args:
        path: Directory relative to the corpus to choose the files from.
        number_of_files: Number of files to select.
        block_size: The block size to be used when fragmenting the files.
        min_fragments: Minimal number of fragments.
        max_fragments: Maximal number of fragments.
    """
    if min_fragments > max_fragments:
        raise WoodblockError('min_fragments has to be <= max_fragments.')
    files = draw_files(path, number_of_files, min_size=block_size * min_fragments)
    frags = list()
    for file in files:
        num_frags = random.randint(min_fragments, min(max_fragments, file.max_fragments(block_size)))  # nosec
        frags.append(file.fragment_randomly(num_frags, block_size))
    return frags


def intertwine_randomly(path: pathlib.Path = None, number_of_files: int = 2, block_size: int = 512,
                        min_fragments: int = 1, max_fragments: int = 4) -> list:
    """Choose random files from ``path`` and intertwine them randomly.

    ``intertwine_randomly`` chooses ``number_of_files`` files from the given ``path``, fragments them at random
    fragmentation points, and intertwines them randomly. The number of fragments per file will be between
    ``min_fragments`` and ``max_fragments`` (both numbers included, i.e. min_fragments <= number of fragments
    <= max_fragments).

    The result is a list of fragments, e.g. [f3.1, f1.1, f3.2, f2.1, f1.2, f2.2, f1.3].

    The function ensures that there will in fact be unique ``number_of_files`` files. Moreover, the function
    guarantees that two fragments of the same file will not be at consecutive list positions. Finally, the fragments
    of each file will be stored in order.

    Args:
        path: Directory to choose the files from.
        number_of_files: Number of files to select.
        block_size: The block size to be used when fragmenting the files.
        min_fragments: Minimal number of fragments.
        max_fragments: Maximal number of fragments.
    """
    if min_fragments > max_fragments:
        raise WoodblockError('min_fragments has to be <= max_fragments.')
    if number_of_files < 2:
        raise WoodblockError('Number of files has to be at least 2.')

    files = draw_files(path, number_of_files, unique=True, min_size=block_size * min_fragments)

    if min_fragments == max_fragments:
        frags = list(itertools.chain.from_iterable(zip(files[0].fragment_randomly(min_fragments, block_size),
                                                       files[1].fragment_randomly(min_fragments, block_size))))
        for file in files[2:]:
            current_file_frags = file.fragment_randomly(min_fragments, block_size)
            _insert_at_slots(frags, current_file_frags, _select_free_slots(len(frags) + 1, min_fragments))
        return frags
    files = sorted(files, key=attrgetter('size'))
    frags_file_1 = _fragment_first_file(files[0], block_size, min_fragments, max_fragments)
    frags_file_2 = _fragment_second_file(files[1], block_size, min_fragments, max_fragments, len(frags_file_1))
    frags = _merge_fragments_of_first_two_files(frags_file_1, frags_file_2)
    min_frags = max(min_fragments, 1)
    for file in files[2:]:
        free_slots = len(frags) + 1
        max_frags = min(max_fragments, file.max_fragments(block_size), free_slots)
        num_frags = random.randint(min_frags, max_frags)  # nosec
        current_file_frags = list(reversed(file.fragment_randomly(num_frags, block_size)))
        _insert_at_slots(frags, current_file_frags, _select_free_slots(free_slots, num_frags))
    return frags


def _merge_fragments_of_first_two_files(frags_file_1, frags_file_2):  # pylint: disable=invalid-name
    frags = frags_file_1[:]
    frags_2 = list(reversed(frags_file_2))
    start = _get_start_index_for_fragments_of_second_file(len(frags_file_1), len(frags_file_2))
    for i in range(len(frags_2)):
        frags.insert(start + 2 * i, frags_2.pop())
    return frags


def _get_start_index_for_fragments_of_second_file(num_frags_file_1, num_frags_file_2):  # pylint: disable=invalid-name
    if num_frags_file_2 == num_frags_file_1 - 1:
        return 1
    if num_frags_file_2 == num_frags_file_1 + 1:
        return 0
    return random.randint(0, 1)  # nosec


def _fragment_first_file(file_1, block_size, min_fragments, max_fragments):
    num_frags = random.randint(max(min_fragments, 1), min(max_fragments, file_1.max_fragments(block_size)))  # nosec
    return file_1.fragment_randomly(num_frags, block_size)


def _fragment_second_file(file_2, block_size, min_fragments, max_fragments, num_fragments_file_1):
    num_frags = random.randint(max(min_fragments, num_fragments_file_1 - 1),  # nosec
                               min(max_fragments, num_fragments_file_1 + 1, file_2.max_fragments(block_size)))
    return file_2.fragment_randomly(num_frags, block_size)


def _insert_at_slots(frag_list, new_fragments, slots):
    new_frags = list(reversed(new_fragments))
    for i, slot in enumerate(slots):
        frag_list.insert(slot + i, new_frags.pop())


def _select_free_slots(num_slots_available, num_slots):
    return sorted(random.sample(range(num_slots_available), k=num_slots))



from setuptools import setup, find_packages


def long_description():
    with open('README.md') as fh:
        desc = fh.read()
    desc = desc.replace('documentation/docs/', 'https://raw.githubusercontent.com/fkie-cad/woodblock/master/documentation/docs/')
    print(desc)
    return desc


setup(
    name='woodblock',
    version='0.1.3',
    license='MIT',
    author='Fraunhofer FKIE',
    author_email='martin.lambertz@fkie.fraunhofer.de',
    url='https://github.com/fkie-cad/woodblock/',
    packages=find_packages(),
    install_requires=[
        'click',
        'multimethod',
        'numpy',
    ],
    entry_points={
        'console_scripts': [
            'woodblock = woodblock.__main__:main'
        ]
    },
    description='A framework to generate file carving test data',
    long_description=long_description(),
    long_description_content_type='text/markdown',
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ],
    project_urls={
        'Documentation': 'https://fkie-cad.github.io/woodblock/',
        'Source': 'https://github.com/fkie-cad/woodblock/',
        'Tracker': 'https://github.com/fkie-cad/woodblock/issues',
    },
)



"""This module contains data generators."""
import random

from woodblock.random import RandomBytes


class Zeroes:
    """Generates zero bytes."""

    def __call__(self, size):
        return b'\x00' * size

    def __str__(self):
        return 'zeroes'


class Random:
    """Generates random bytes."""

    def __init__(self, rng=RandomBytes()):
        self._seed = random.randint(0, 2 ** 32 - 1)  # nosec
        self._rng = rng
        self._rng.seed(self._seed)

    def __call__(self, size):
        return self._rng.bytes(size)

    def __str__(self):
        return 'random'

    def reset(self):
        """Reset the random state.

        After calling this method, the RNG will be set to its initial seed, so that subsequent calls return the same
        bytes as the calls before.
        """
        self._rng.seed(self._seed)



"""This module contains the different fragment classes which can be used in file carving scenarios."""

import hashlib
from uuid import uuid4

import woodblock.datagen
from woodblock.errors import WoodblockError


class FillerFragment:
    """A filler fragment.

    A filler fragment is a fragment containing synthetic data (e.g. random data). It can be used to simulate wiped
    areas or areas with random data.
    """

    def __init__(self, size, data_generator=None, chunk_size=8192):
        self._size = size
        self._chunk_size = min(size, chunk_size)
        self._orig_chunk_size = chunk_size
        if data_generator is None:
            data_generator = woodblock.datagen.Random()
        self._generate_data = data_generator
        self._bytes_read = 0
        if self._size < 1:
            raise WoodblockError('Fragments must not be empty!')
        self._hash = None
        self._hasher = hashlib.sha256()
        self._id = uuid4().hex

    @property
    def size(self):
        """Return the size of the fragment."""
        return self._size

    @property
    def hash(self):
        """Return the SHA-256 digest as hexadecimal string."""
        if self._hash is None:
            self._hash = self._compute_hash()
        return self._hash

    @property
    def metadata(self) -> dict:
        """Return the fragment metadata."""
        if self._hash is None:
            self._hash = self._compute_hash()
        return {'file': {'type': 'filler', 'sha256': self.hash, 'size': self.size, 'path': str(self._generate_data),
                         'id': self._id},
                'fragment': {'sha256': self.hash, 'size': self.size, 'number': 1,
                             'file_offsets': {'start': 0, 'end': self.size}}}

    def __iter__(self):
        self._hasher = hashlib.sha256()
        return self

    def __next__(self):
        if self._chunk_size <= 0:
            self._reset_state()
            raise StopIteration()
        chunk = self._generate_data(min(self._chunk_size, self._size))
        self._bytes_read += self._chunk_size
        self._chunk_size = max(0, min(self._chunk_size, self._size - self._bytes_read))
        if self._hash is None:
            self._hasher.update(chunk)
        return chunk

    def _compute_hash(self):
        hasher = hashlib.sha256()
        for chunk in self:
            hasher.update(chunk)
        return hasher.hexdigest()

    def _reset_state(self):
        if self._hash is None:
            self._hash = self._hasher.hexdigest()
        self._chunk_size = self._orig_chunk_size


class ZeroesFragment(FillerFragment):
    """A fragment filled completely with zero bytes (0x00)."""

    def __init__(self, size, chunk_size=8192):
        super().__init__(size=size, data_generator=woodblock.datagen.Zeroes(), chunk_size=chunk_size)


class RandomDataFragment(FillerFragment):
    """A fragment filled with random bytes."""

    def __init__(self, size, chunk_size=8192):
        self._rng = woodblock.datagen.Random()
        super().__init__(size=size, data_generator=self._rng, chunk_size=chunk_size)

    def _reset_state(self):
        super()._reset_state()
        self._rng.reset()


class FileFragment:
    """A fragment of an actual file."""

    def __init__(self, file, fragment_number, start_offset, end_offset, chunk_size=8192):
        self._file = file
        self._number = fragment_number
        self._start_offset = start_offset
        self._end_offset = end_offset
        self._size = end_offset - start_offset
        self._hash = None
        self._hasher = hashlib.sha256()
        self._chunk_size = chunk_size
        self._orig_chunk_size = chunk_size
        self._bytes_read = 0
        self._file_handle = None

    @property
    def size(self):
        """Return the size of the fragment."""
        return self._size

    @property
    def hash(self):
        """Return the SHA-256 digest as hexadecimal string."""
        if self._hash is None:
            self._compute_hash()
        return self._hash

    def __iter__(self):
        self._hasher = hashlib.sha256()
        self._file_handle = open(self._file.path, 'rb')
        self._file_handle.seek(self._start_offset)
        return self

    def __next__(self):
        if self._chunk_size <= 0:
            self._reset_state()
            raise StopIteration()
        chunk = self._file_handle.read(min(self._chunk_size, self._size))
        self._bytes_read += len(chunk)
        self._chunk_size = max(0, min(self._chunk_size, self._size - self._bytes_read))
        if self._hash is None:
            self._hasher.update(chunk)
        return chunk

    @property
    def metadata(self):
        """Return the fragment metadata."""
        if self._hash is None:
            self._compute_hash()
        return {'file': {'type': 'file', 'sha256': self._file.hash, 'size': self._file.size,
                         'path': str(self._file.path.relative_to(woodblock.file.get_corpus())), 'id': self._file.id},
                'fragment': {'sha256': self.hash, 'size': self.size, 'number': self._number,
                             'file_offsets': {'start': self._start_offset, 'end': self._end_offset}}}

    def _compute_hash(self):
        for _ in self:
            pass

    def _reset_state(self):
        if self._hash is None:
            self._hash = self._hasher.hexdigest()
        self._file_handle.close()
        self._chunk_size = self._orig_chunk_size
